"""
Tests for analyze module

Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan

Distributed under the terms of the MIT License

@author: Andrew R. McCluskey
"""

import unittest
import numpy as np
from numpy.testing import assert_equal, assert_almost_equal, assert_raises
from kinisi.matrix import find_nearest_positive_definite, check_positive_definite


class TestFunctions(unittest.TestCase):
    """
    Tests for the matrix functions.
    """

    def test_find_psd_matrix(self):
        start_matrix = np.eye(20)
        assert_equal(start_matrix, find_nearest_positive_definite(start_matrix))

    def test_find_non_psd_matrix(self):
        start_matrix = np.array([[0.99, 0.78, 0.59, 0.44], [0.78, 0.92, 0.28, 0.81], [0.59, 0.28, 1.12, 0.23],
                                 [0.44, 0.81, 0.23, 0.99]])
        result = find_nearest_positive_definite(start_matrix)
        assert_raises(AssertionError, assert_almost_equal, start_matrix, result)

    def test_check_psd_matrix(self):
        start_matrix = np.eye(20)
        assert check_positive_definite(start_matrix) is True

    def test_check_non_psd_matrix(self):
        start_matrix = np.array([[0.99, 0.78, 0.59, 0.44], [0.78, 0.92, 0.28, 0.81], [0.59, 0.28, 1.12, 0.23],
                                 [0.44, 0.81, 0.23, 0.99]])
        assert check_positive_definite(start_matrix) is False



"""
This module contains the API classes for :py:mod:`kinisi`.
It is anticipated that this is where the majority of interaction with the package will occur.
This module includes:

* the :py:class:`kinisi.analyze.DiffusionAnalyzer` class for MSD and diffusion analysis;
* the :py:class:`kinisi.analyze.JumpDiffusionAnalyzer` class for TMSD and collective diffusion analysis; and
* the :py:class:`kinisi.analyze.ConductivityAnalyzer` class for MSCD and conductivity analysis.

These are all compatible with VASP Xdatcar output files, pymatgen structures and any MD trajectory that the
:py:mod:`MDAnalysis` package can handle.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (arm61)

from .diffusion_analyzer import DiffusionAnalyzer
from .jump_diffusion_analyzer import JumpDiffusionAnalyzer
from .conductivity_analyzer import ConductivityAnalyzer



"""
Tests for arrhenius module

Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan

Distributed under the terms of the MIT License

@author: Andrew R. McCluskey
"""

# pylint: disable=R0201

import unittest
import numpy as np
from scipy.stats import norm
from numpy.testing import assert_equal, assert_almost_equal
from kinisi import arrhenius
from uravu.distribution import Distribution

temp = np.linspace(5, 50, 10)
ea = np.linspace(5, 50, 10)
EA = []
for e in ea:
    EA.append(Distribution(norm.rvs(loc=e, scale=e * 0.1, size=5000, random_state=np.random.RandomState(1))))


class TestArrhenius(unittest.TestCase):
    """
    Unit tests for arrhenius module
    """

    def test_standard_arrhenius_init(self):
        """
        Test the initialisation of standard arrhenius
        """
        arr = arrhenius.StandardArrhenius(temp, EA, ((0, 1000), (0, 1000)))
        assert_equal(arr.function, arrhenius.arrhenius)
        assert_almost_equal(arr.abscissa, temp)
        assert_almost_equal(arr.y.n, ea, decimal=0)
        assert_almost_equal(arr.y.s, np.array([ea * 0.196, ea * 0.196]), decimal=0)

    def test_super_arrhenius_init(self):
        """
        Test the initialisation of super arrhenius
        """
        arr = arrhenius.SuperArrhenius(temp, EA, ((0, 1000), (0, 1000), (0, 4)))
        assert_equal(arr.function, arrhenius.super_arrhenius)
        assert_almost_equal(arr.abscissa, temp)
        assert_almost_equal(arr.y.n, ea, decimal=0)
        assert_almost_equal(arr.y.s, np.array([ea * 0.196, ea * 0.196]), decimal=0)

    def test_standard_arrhenius(self):
        """
        Test the arrhenius function
        """
        assert_almost_equal(9.996132574, arrhenius.arrhenius(300, 1e-5, 10), decimal=5)

    def test_super_arrhenius(self):
        """
        Test the super arrhenius function
        """
        assert_almost_equal(9.995999241, arrhenius.super_arrhenius(300, 1e-5, 10, 10), decimal=5)



"""
The :py:class:`kinisi.analyze.JumpDiffusionAnalyzer` class allows the study of jump diffusion
and the collective motion of particles.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (arm61)

from typing import Union, List
import numpy as np
from kinisi import diffusion
from .analyzer import Analyzer


class JumpDiffusionAnalyzer(Analyzer):
    """
    The :py:class:`kinisi.analyze.JumpDiffusionAnalyzer` class performs analysis of collective diffusion
    relationships in materials.
    This is achieved through the application of a bootstrapping methodology to obtain the most statistically
    accurate values for total mean squared displacement uncertainty and covariance.
    The time-dependence of the TMSD is then modelled in a generalised least squares fashion to obtain the jump
    diffusion coefficient and offset using Markov chain Monte Carlo maximum likelihood sampling.

    :param delta_t: An array of the timestep values.
    :param disp_3d: A list of arrays, where each array has the axes [atom, displacement observation, dimension].
        There is one array in the list for each delta_t value. Note: it is necessary to use a list of arrays as
        the number of observations is not necessary the same at each data point.
    :param volume: The volume of the simulation cell.
    :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.DiffBootstrap` object. See
        the appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.
    """

    def __init__(self, delta_t: np.ndarray, disp_3d: List[np.ndarray], volume: float):
        super().__init__(delta_t, disp_3d, volume)
        self._diff = None

    def to_dict(self) -> dict:
        """
        :return: Dictionary description of :py:class:`JumpDiffusionAnalyzer`.
        """
        my_dict = super().to_dict()
        my_dict['diff'] = self._diff.to_dict()
        return my_dict

    @classmethod
    def from_dict(cls, my_dict) -> 'JumpDiffusionAnalyzer':
        """
        Generate a :py:class:`DiffusionAnalyzer` object from a dictionary.

        :param my_dict: The input dictionary.

        :return: New :py:class:`DiffusionAnalyzer` object.
        """
        jdiff_anal = cls(my_dict['delta_t'], my_dict['disp_3d'], my_dict['volume'])
        jdiff_anal._diff = diffusion.Bootstrap.from_dict(my_dict['diff'])
        return jdiff_anal

    @classmethod
    def from_pymatgen(cls,
                      trajectory: List[Union['pymatgen.core.structure.Structure',
                                             List['pymatgen.core.structure.Structure']]],
                      parser_params: dict,
                      dtype: str = None,
                      bootstrap_params: dict = None):
        """
        Create a :py:class:`JumpDiffusionAnalyzer` object from a list or nested list of
        :py:class:`pymatgen.core.structure.Structure` objects.

        :param trajectory: The list or nested list of structures to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a list of :py:class:`pymatgen.core.structure.Structure` objects,
            this should be :py:attr:`None`. However, if a list of lists is passed, then it is necessary to identify if
            these constitute a series of :py:attr:`consecutive` trajectories or a series of :py:attr:`identical`
            starting points with different random seeds, in which case the `dtype` should be either
            :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`JumpDiffusionAnalyzer` object.
        """
        if bootstrap_params is None:
            bootstrap_params = {}
        jdiff_anal = super()._from_pymatgen(trajectory, parser_params, dtype=dtype)
        jdiff_anal._diff = diffusion.TMSDBootstrap(jdiff_anal._delta_t, jdiff_anal._disp_3d, **bootstrap_params)
        return jdiff_anal

    @classmethod
    def from_Xdatcar(cls,
                     trajectory: Union['pymatgen.io.vasp.outputs.Xdatcar', List['pymatgen.io.vasp.outputs.Xdatcar']],
                     parser_params: dict,
                     dtype: str = None,
                     bootstrap_params: dict = None):
        """
        Create a :py:class:`JumpDiffusionAnalyzer` object from a single or a list of
        :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects.

        :param trajectory: The Xdatcar or list of Xdatcar objects to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a :py:class:`pymatgen.io.vasp.outputs.Xdatcar` object, this should
            be :py:attr:`None`. However, if a list of :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects is passed,
            then it is necessary to identify if these constitute a series of :py:attr:`consecutive` trajectories or a
            series of :py:attr:`identical` starting points with different random seeds, in which case the `dtype`
            should be either :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`JumpDiffusionAnalyzer` object.
        """
        if bootstrap_params is None:
            bootstrap_params = {}
        jdiff_anal = super()._from_Xdatcar(trajectory, parser_params, dtype=dtype)
        jdiff_anal._diff = diffusion.TMSDBootstrap(jdiff_anal._delta_t, jdiff_anal._disp_3d, **bootstrap_params)
        return jdiff_anal

    @classmethod
    def from_file(cls,
                  trajectory: Union[str, List[str]],
                  parser_params: dict,
                  dtype: str = None,
                  bootstrap_params: dict = None):
        """
        Create a :py:class:`JumpDiffusionAnalyzer` object from a single or a list of Xdatcar file(s).

        :param trajectory: The file or list of Xdatcar files to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify if these constitute a series of
            :py:attr:`consecutive` trajectories or a series of :py:attr:`identical` starting points with different
            random seeds, in which case the `dtype` should be either :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`JumpDiffusionAnalyzer` object.
        """
        if bootstrap_params is None:
            bootstrap_params = {}
        jdiff_anal = super()._from_file(trajectory, parser_params, dtype=dtype)
        jdiff_anal._diff = diffusion.TMSDBootstrap(jdiff_anal._delta_t, jdiff_anal._disp_3d, **bootstrap_params)
        return jdiff_anal

    @classmethod
    def from_universe(cls,
                      trajectory: 'MDAnalysis.core.universe.Universe',
                      parser_params: dict,
                      dtype: str = None,
                      bootstrap_params: dict = None):
        """
        Create an :py:class:`JumpDiffusionAnalyzer` object from an :py:class:`MDAnalysis.core.universe.Universe` object.

        :param trajectory: The universe to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.MDAnalysisParser` object.
            See the appropriate documention for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify that these constitute a series of
            :py:attr:`identical` starting points with different random seeds, in which case the `dtype` should
            be :py:attr:`identical`. For a series of consecutive trajectories, please construct the relevant
            object using :py:mod:`MDAnalysis`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`JumpDiffusionAnalyzer` object.
        """
        if bootstrap_params is None:
            bootstrap_params = {}
        jdiff_anal = super()._from_universe(trajectory, parser_params, dtype=dtype)
        jdiff_anal._diff = diffusion.TMSDBootstrap(jdiff_anal._delta_t, jdiff_anal._disp_3d, **bootstrap_params)
        return jdiff_anal

    def jump_diffusion(self, jump_diffusion_params: Union[dict, None] = None):
        """
        Calculate the jump diffusion coefficicent using the bootstrap-GLS methodology.

        :param ump_diffusion_params: The parameters for the :py:class:`kinisi.diffusion.TMSDBootstrap`
            object. See the appropriate documentation for more guidance on this. Optional, default is the
            default bootstrap parameters.
        """
        if jump_diffusion_params is None:
            jump_diffusion_params = {}
        self._diff.jump_diffusion(**jump_diffusion_params)

    @property
    def tmsd(self) -> np.ndarray:
        """
        :return: TMSD for the input trajectories. Note that this is the bootstrap sampled MSD, not the numerical
            average from the data.
        """
        return self._diff.n

    @property
    def tmsd_std(self) -> np.ndarray:
        """
        :return: MSD standard deviations values for the input trajectories.
        """
        return self._diff.s

    @property
    def D_J(self) -> 'uravu.distribution.Distribution':
        """
        :return: Jump diffusion coefficient
        """
        return self._diff.D_J

    @property
    def flatchain(self) -> np.ndarray:
        """
        :return: sampling flatchain
        """
        return np.array([self.D_J.samples, self.intercept.samples]).T



"""
Tests for diffusion module

Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan

Distributed under the terms of the MIT License

@author: Andrew R. McCluskey
"""

# pylint: disable=R0201

import unittest
import warnings
import numpy as np
from tqdm import tqdm
from numpy.testing import assert_almost_equal, assert_equal
from kinisi.diffusion import Bootstrap, MSDBootstrap, TMSDBootstrap, MSCDBootstrap, _bootstrap
from uravu.distribution import Distribution

RNG = np.random.RandomState(42)


class TestBootstrap(unittest.TestCase):
    """
    Tests for the diffusion.Bootstrap class.
    """

    def test_dictionary_roundtrip(self):
        disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
        dt = np.linspace(100, 1000, 10)
        a = Bootstrap(dt, disp_3d)
        b = Bootstrap.from_dict(a.to_dict())
        for i, d in enumerate(disp_3d):
            assert_almost_equal(a._displacements[i], b._displacements[i])
        assert_almost_equal(a._delta_t, b._delta_t)
        assert a._max_obs == b._max_obs
        assert a._distributions == b._distributions
        assert a.dt.size == b.dt.size
        assert isinstance(b.dt, np.ndarray)
        assert b._diffusion_coefficient is None
        assert b.intercept is None

    def test_initialisation(self):
        disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
        dt = np.linspace(100, 1000, 10)
        bs = Bootstrap(dt, disp_3d)
        for i, d in enumerate(disp_3d):
            assert_almost_equal(bs._displacements[i], d)
        assert_almost_equal(bs._delta_t, dt)
        assert bs._max_obs == 20
        assert bs._distributions == []
        assert bs.dt.size == 0
        assert isinstance(bs.dt, np.ndarray)
        assert isinstance(bs._iterator, tqdm)
        assert bs._diffusion_coefficient is None
        assert bs.intercept is None

    def test_initialisation_sub_sample_dt(self):
        disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
        dt = np.linspace(100, 1000, 10)
        bs = Bootstrap(dt, disp_3d, sub_sample_dt=2)
        for i, d in enumerate(disp_3d[::2]):
            assert_almost_equal(bs._displacements[i], d)
        assert_almost_equal(bs._delta_t, dt[::2])
        assert bs._max_obs == 20
        assert bs._distributions == []
        assert bs.dt.size == 0
        assert isinstance(bs.dt, np.ndarray)
        assert isinstance(bs._iterator, tqdm)
        assert bs._diffusion_coefficient is None
        assert bs.intercept is None

    def test_initialisation_progress(self):
        disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
        dt = np.linspace(100, 1000, 10)
        bs = Bootstrap(dt, disp_3d, progress=False)
        for i, d in enumerate(disp_3d):
            assert_almost_equal(bs._displacements[i], d)
        assert_almost_equal(bs._delta_t, dt)
        assert bs._max_obs == 20
        assert bs._distributions == []
        assert bs.dt.size == 0
        assert isinstance(bs.dt, np.ndarray)
        assert isinstance(bs._iterator, range)
        assert bs._diffusion_coefficient is None
        assert bs.intercept is None

    def test_iterator_true(self):
        result = Bootstrap.iterator(True, range(10))
        assert isinstance(result, tqdm)

    def test_iterator_false(self):
        result = Bootstrap.iterator(False, range(10))
        assert isinstance(result, range)

    def test_n_samples_a(self):
        result = Bootstrap.n_samples((100, 10), 10)
        assert result == 1000

    def test_n_samples_b(self):
        result = Bootstrap.n_samples((100, 8), 10)
        assert result == 333

    def test_sample_until_normal(self):
        distro1 = Bootstrap.sample_until_normal(RNG.normal(0, 1, size=2000),
                                                50,
                                                1000,
                                                100000,
                                                random_state=np.random.RandomState(1))
        assert distro1.size == 1000

    def test_sample_until_normal_random(self):
        with warnings.catch_warnings(record=True) as _:
            distro1 = Bootstrap.sample_until_normal(np.arange(1, 10, 1),
                                                    5,
                                                    100,
                                                    100,
                                                    random_state=np.random.RandomState(0))
            distro2 = Bootstrap.sample_until_normal(np.arange(1, 10, 1),
                                                    5,
                                                    100,
                                                    100,
                                                    random_state=np.random.RandomState(0))
            assert_almost_equal(distro1.samples, distro2.samples)

    def test_ngp_calculation(self):
        result = Bootstrap.ngp_calculation(np.array([1, 2, 3]))
        assert_almost_equal(result, -0.3)


class TestMSDBootstrap(unittest.TestCase):
    """
    Tests for the diffusion.MSDBootstrap class.
    """

    def test_dictionary_roundtrip(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            a = MSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
            b = MSDBootstrap.from_dict(a.to_dict())
            for i, d in enumerate(disp_3d):
                assert_almost_equal(a._displacements[i], b._displacements[i])
            assert_almost_equal(a._delta_t, b._delta_t)
            assert a._max_obs == b._max_obs
            assert_equal(a.n, b.n)
            assert_equal(a.s, b.s)
            assert_equal(a.v, b.v)
            assert_equal(a.n_i, b.n_i)
            assert_equal(a.ngp, b.ngp)
            for i, d in enumerate(a._distributions):
                assert_almost_equal(d.samples, b._distributions[i].samples)
            for i, d in enumerate(a.euclidian_displacements):
                assert_equal(d.samples, b.euclidian_displacements[i].samples)
            assert a.dt.size == b.dt.size
            assert isinstance(b.dt, np.ndarray)
            assert b._diffusion_coefficient is None
            assert b.intercept is None

    def test_initialisation(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [1000, 666, 500, 400, 333, 285, 250, 222, 200])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])

    def test_initialisation_n_resamples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, n_resamples=10, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [1000, 666, 500, 400, 333, 285, 250, 222, 200])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 10
                assert_almost_equal(i.ci_points, [2.5, 97.5])

    def test_initialisation_max_resamples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, n_resamples=10, max_resamples=100, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [1000, 666, 500, 400, 333, 285, 250, 222, 200])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size <= 110
                assert_almost_equal(i.ci_points, [2.5, 97.5])

    def test_initialisation_random_state(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs1 = MSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
            assert bs1.n.shape == (9, )
            assert bs1.s.shape == (9, )
            assert_almost_equal(bs1.v, np.square(bs1.s))
            assert_equal(bs1.n_i, [1000, 666, 500, 400, 333, 285, 250, 222, 200])
            assert bs1.ngp.shape == (9, )
            assert len(bs1.euclidian_displacements) == 9
            for i in bs1.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs1._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])
            bs2 = MSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
            assert bs1._distributions[-1].size == bs2._distributions[-1].size
            assert_almost_equal(bs1._distributions[-1].samples, bs2._distributions[-1].samples)

    def test_initialisation_progress(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, progress=False, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [1000, 666, 500, 400, 333, 285, 250, 222, 200])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])
            assert isinstance(bs._iterator, range)

    def test_initialisation_skip_where_low_samples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(1, i, 3) for i in range(10, 1, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, progress=False, random_state=np.random.RandomState(0))
            assert bs.n.shape == (4, )
            assert bs.s.shape == (4, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [5, 3, 2, 2])
            assert bs.ngp.shape == (4, )
            assert len(bs.euclidian_displacements) == 4
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])
            assert isinstance(bs._iterator, range)

    def test_bootstrap_dictionary_roundtrip(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            a = MSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
            a.diffusion()
            b = MSDBootstrap.from_dict(a.to_dict())
            for i, d in enumerate(disp_3d):
                assert_almost_equal(a._displacements[i], b._displacements[i])
            assert_almost_equal(a._delta_t, b._delta_t)
            assert a._max_obs == b._max_obs
            assert_equal(a.n, b.n)
            assert_equal(a.s, b.s)
            assert_equal(a.v, b.v)
            assert_equal(a.n_i, b.n_i)
            assert_equal(a.ngp, b.ngp)
            for i, d in enumerate(a._distributions):
                assert_almost_equal(d.samples, b._distributions[i].samples)
            for i, d in enumerate(a.euclidian_displacements):
                assert_equal(d.samples, b.euclidian_displacements[i].samples)
            assert a.dt.size == b.dt.size
            assert isinstance(b.dt, np.ndarray)
            assert_equal(a._diffusion_coefficient.samples, b._diffusion_coefficient.samples)
            assert_equal(a.intercept.samples, b.intercept.samples)

    def test_bootstrap(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.diffusion()
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs._diffusion_coefficient, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs._diffusion_coefficient.size == 32000
            assert bs.intercept.size == 32000

    def test_bootstrap_dt_skip(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d)
            bs.diffusion(dt_skip=150)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs._diffusion_coefficient, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs._diffusion_coefficient.size == 32000
            assert bs.intercept.size == 32000

    def test_bootstrap_use_ngp(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(200, 190, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.diffusion(use_ngp=True)
            assert bs.covariance_matrix.shape == (9 - np.argmax(bs.ngp), 9 - np.argmax(bs.ngp))
            assert isinstance(bs._diffusion_coefficient, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs._diffusion_coefficient.size == 32000
            assert bs.intercept.size == 32000

    def test_bootstrap_fit_intercept(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.diffusion(n_samples=500, fit_intercept=False)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs._diffusion_coefficient, Distribution)
            assert bs._diffusion_coefficient.size == 16000
            assert bs.intercept is None

    def test_bootstrap_n_samples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.diffusion(n_samples=100)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs._diffusion_coefficient, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs._diffusion_coefficient.size == 3200
            assert bs.intercept.size == 3200

    def test_bootstrap_D(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.diffusion(n_samples=100)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs._diffusion_coefficient, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs.D.size == 3200
            assert bs.intercept.size == 3200

    # Waiting on https://github.com/dfm/emcee/pull/376
    # def test_initialisation_random_state(self):
    #     disp_3d = [RNG.randn(100, i, 3) + 1000 for i in range(20, 10, -1)]
    #     dt = np.linspace(100, 1000, 10)
    #     bs1 = DiffBootstrap(dt, disp_3d, n_walkers=5, n_samples=100, random_state=np.random.RandomState(0))
    #     assert bs1.covariance_matrix.shape == (10, 10)
    #     assert isinstance(bs1.diffusion_coefficient, Distribution)
    #     assert isinstance(bs1.intercept, Distribution)
    #     assert bs1.diffusion_coefficient.size == 500
    #     assert bs1.intercept.size == 500
    #     bs2 = DiffBootstrap(dt, disp_3d, n_walkers=5, n_samples=100, random_state=np.random.RandomState(0))
    #     assert_almost_equal(bs1.v, bs2.v)
    #     assert_almost_equal(bs1.covariance_matrix, bs2.covariance_matrix)
    #     assert_almost_equal(bs1.diffusion_coefficient.samples, bs2.diffusion_coefficient.samples)


class TestTMSDBootstrap(unittest.TestCase):
    """
    Tests for the diffusion.TMSDBootstrap class.
    """

    def test_initialisation(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])

    def test_initialisation_n_resamples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, n_resamples=10, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 10
                assert_almost_equal(i.ci_points, [2.5, 97.5])

    def test_initialisation_max_resamples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, n_resamples=10, max_resamples=100, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size <= 110
                assert_almost_equal(i.ci_points, [2.5, 97.5])

    def test_initialisation_random_state(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs1 = TMSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
            assert bs1.n.shape == (9, )
            assert bs1.s.shape == (9, )
            assert_almost_equal(bs1.v, np.square(bs1.s))
            assert_equal(bs1.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs1.ngp.shape == (9, )
            assert len(bs1.euclidian_displacements) == 9
            for i in bs1.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs1._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])
            bs2 = TMSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
            assert bs1._distributions[-1].size == bs2._distributions[-1].size
            assert_almost_equal(bs1._distributions[-1].samples, bs2._distributions[-1].samples)

    def test_initialisation_progress(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, progress=False, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])
            assert isinstance(bs._iterator, range)

    def test_initialisation_skip_where_low_samples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(1, i, 3) for i in range(10, 1, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, progress=False, random_state=np.random.RandomState(0))
            assert bs.n.shape == (4, )
            assert bs.s.shape == (4, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [5, 3, 2, 2])
            assert bs.ngp.shape == (4, )
            assert len(bs.euclidian_displacements) == 4
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])
            assert isinstance(bs._iterator, range)

    def test_bootstrap(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.jump_diffusion()
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs._jump_diffusion_coefficient, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs._jump_diffusion_coefficient.size == 32000
            assert bs.intercept.size == 32000

    def test_bootstrap_use_ngp(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.jump_diffusion(use_ngp=True)
            assert bs.covariance_matrix.shape == (9 - np.argmax(bs.ngp), 9 - np.argmax(bs.ngp))
            assert isinstance(bs._jump_diffusion_coefficient, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs._jump_diffusion_coefficient.size == 32000
            assert bs.intercept.size == 32000

    def test_bootstrap_fit_intercept(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.jump_diffusion(n_samples=500, fit_intercept=False)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs._jump_diffusion_coefficient, Distribution)
            assert bs._jump_diffusion_coefficient.size == 16000
            assert bs.intercept is None

    def test_bootstrap_n_samples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.jump_diffusion(n_samples=100)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs._jump_diffusion_coefficient, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs._jump_diffusion_coefficient.size == 3200
            assert bs.intercept.size == 3200

    def test_bootstrap_D(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = TMSDBootstrap(dt, disp_3d, random_state=RNG)
            bs.jump_diffusion(n_samples=100)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs._jump_diffusion_coefficient, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs._jump_diffusion_coefficient.size == 3200
            assert bs.intercept.size == 3200

    # Waiting on https://github.com/dfm/emcee/pull/376
    # def test_initialisation_random_state(self):
    #     disp_3d = [RNG.randn(100, i, 3) + 1000 for i in range(20, 10, -1)]
    #     dt = np.linspace(100, 1000, 10)
    #     bs1 = TMSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
    #     bs1.jump_diffusion( n_walkers=5, n_samples=100, random_state=np.random.RandomState(0))
    #     assert bs1.covariance_matrix.shape == (10, 10)
    #     assert isinstance(bs1.diffusion_coefficient, Distribution)
    #     assert isinstance(bs1.intercept, Distribution)
    #     assert bs1.diffusion_coefficient.size == 500
    #     assert bs1.intercept.size == 500
    #     bs2 = TMSDBootstrap(dt, disp_3d, random_state=np.random.RandomState(0))
    #     bs2.jump_diffusion( n_walkers=5, n_samples=100, random_state=np.random.RandomState(0))
    #     assert_almost_equal(bs1.v, bs2.v)
    #     assert_almost_equal(bs1.covariance_matrix, bs2.covariance_matrix)
    #     assert_almost_equal(bs1.diffusion_coefficient.samples, bs2.diffusion_coefficient.samples)


class TestMSCDBootstrap(unittest.TestCase):
    """
    Tests for the diffusion.MSCDBootstrap class.
    """

    def test_initialisation(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, 1, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])

    def test_initialisation_n_resamples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, np.ones(100), n_resamples=10, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 10
                assert_almost_equal(i.ci_points, [2.5, 97.5])

    def test_initialisation_max_resamples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, 1, n_resamples=10, max_resamples=100, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size <= 110
                assert_almost_equal(i.ci_points, [2.5, 97.5])

    def test_initialisation_random_state(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs1 = MSCDBootstrap(dt, disp_3d, 1, random_state=np.random.RandomState(0))
            assert bs1.n.shape == (9, )
            assert bs1.s.shape == (9, )
            assert_almost_equal(bs1.v, np.square(bs1.s))
            assert_equal(bs1.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs1.ngp.shape == (9, )
            assert len(bs1.euclidian_displacements) == 9
            for i in bs1.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs1._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])
            bs2 = MSCDBootstrap(dt, disp_3d, 1, random_state=np.random.RandomState(0))
            assert bs1._distributions[-1].size == bs2._distributions[-1].size
            assert_almost_equal(bs1._distributions[-1].samples, bs2._distributions[-1].samples)

    def test_initialisation_progress(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, 1, progress=False, random_state=np.random.RandomState(0))
            assert bs.n.shape == (9, )
            assert bs.s.shape == (9, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [10, 6, 5, 4, 3, 2, 2, 2, 2])
            assert bs.ngp.shape == (9, )
            assert len(bs.euclidian_displacements) == 9
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])
            assert isinstance(bs._iterator, range)

    def test_initialisation_skip_where_low_samples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(1, i, 3) for i in range(10, 1, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, 1, progress=False, random_state=np.random.RandomState(0))
            assert bs.n.shape == (4, )
            assert bs.s.shape == (4, )
            assert_almost_equal(bs.v, np.square(bs.s))
            assert_equal(bs.n_i, [5, 3, 2, 2])
            assert bs.ngp.shape == (4, )
            assert len(bs.euclidian_displacements) == 4
            for i in bs.euclidian_displacements:
                assert isinstance(i, Distribution)
            for i in bs._distributions:
                assert i.samples.size >= 1000
                assert_almost_equal(i.ci_points, [2.5, 97.5])
            assert isinstance(bs._iterator, range)

    def test_bootstrap(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, 1, random_state=RNG)
            bs.conductivity(1, 10)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs.sigma, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs.sigma.size == 32000
            assert bs.intercept.size == 32000

    def test_bootstrap_use_ngp(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, 1, random_state=RNG)
            bs.conductivity(1, 10, use_ngp=True)
            assert bs.covariance_matrix.shape == (9 - np.argmax(bs.ngp), 9 - np.argmax(bs.ngp))
            assert isinstance(bs.sigma, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs.sigma.size == 32000
            assert bs.intercept.size == 32000

    def test_bootstrap_fit_intercept(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, 1, random_state=RNG)
            bs.conductivity(1, 10, n_samples=500, fit_intercept=False)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs.sigma, Distribution)
            assert bs.sigma.size == 16000
            assert bs.intercept is None

    def test_bootstrap_n_samples(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, 1, random_state=RNG)
            bs.conductivity(1, 10, n_samples=100)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs.sigma, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs.sigma.size == 3200
            assert bs.intercept.size == 3200

    def test_bootstrap_D(self):
        with warnings.catch_warnings(record=True) as _:
            disp_3d = [RNG.randn(100, i, 3) for i in range(20, 10, -1)]
            dt = np.linspace(100, 1000, 10)
            bs = MSCDBootstrap(dt, disp_3d, 1, random_state=RNG)
            bs.conductivity(1, 10, n_samples=100)
            assert bs.covariance_matrix.shape == (9, 9)
            assert isinstance(bs.sigma, Distribution)
            assert isinstance(bs.intercept, Distribution)
            assert bs.sigma.size == 3200
            assert bs.intercept.size == 3200

    # Waiting on https://github.com/dfm/emcee/pull/376
    # def test_initialisation_random_state(self):
    #     disp_3d = [RNG.randn(100, i, 3) + 1000 for i in range(20, 10, -1)]
    #     dt = np.linspace(100, 1000, 10)
    #     bs1 = DiffBootstrap(dt, disp_3d, n_samples=100, random_state=np.random.RandomState(0))
    #     assert bs1.covariance_matrix.shape == (10, 10)
    #     assert isinstance(bs1.diffusion_coefficient, Distribution)
    #     assert isinstance(bs1.intercept, Distribution)
    #     assert bs1.diffusion_coefficient.size == 500
    #     assert bs1.intercept.size == 500
    #     bs2 = DiffBootstrap(dt, disp_3d, n_samples=100, random_state=np.random.RandomState(0))
    #     assert_almost_equal(bs1.v, bs2.v)
    #     assert_almost_equal(bs1.covariance_matrix, bs2.covariance_matrix)
    #     assert_almost_equal(bs1.diffusion_coefficient.samples, bs2.diffusion_coefficient.samples)


class TestFunctions(unittest.TestCase):
    """
    Testing other functions.
    """

    def test__bootstrap_random(self):
        result1 = _bootstrap(np.arange(1, 100, 1), 200, 100, np.random.RandomState(0))
        result2 = _bootstrap(np.arange(1, 100, 1), 200, 100, np.random.RandomState(0))
        assert_almost_equal(result1, result2)



"""
This module contains the base class for the different :py:class:`Analyzer` objects used by :py:mod:`kinisi`.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (arm61)

from typing import Union, List
import numpy as np
from kinisi.parser import MDAnalysisParser, PymatgenParser


class Analyzer:
    """
    This class is the superclass for the :py:class:`kinisi.analyze.DiffusionAnalyzer`,
    :py:class:`kinisi.analyze.JumpDiffusionAnalyzer` and :py:class:`kinisi.analyze.ConductivityAnalyzer` classes.
    Therefore all of the properties here are available to these other classes.

    :param delta_t: An array of the timestep values.
    :param disp_3d: A list of arrays, where each array has the axes :code:`[atom, displacement observation,
        dimension]`. There is one array in the list for each delta_t value. Note: it is necessary to use a
        list of arrays as the number of observations is not necessary the same at each data point.
    :param volume: The volume of the simulation cell.
    """

    def __init__(self, delta_t: np.ndarray, disp_3d: List[np.ndarray], volume: float):
        self._delta_t = delta_t
        self._disp_3d = disp_3d
        self._volume = volume

    def save(self, filename: str):
        """
        Save the Analyzer object as a HDF5 file.

        :param filename: Name for the file, no file extension is required and if one if given it is replaced with .hdf.
        """
        if filename[-4:] != '.hdf':
            raise ValueError("The expected file extension is .hdf")
        my_dict = self.to_dict()
        try:
            import h5py
            from os.path import exists
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError("To save and load objects, the h5py module is required")  # pragma: no cover
        if exists(filename):
            raise ValueError(f"The file {filename} already exists, please delete it or change the input filename.")
        with h5py.File(filename, 'w') as h5file:
            _dict_to_group(h5file, '/', my_dict)

    @classmethod
    def load(cls, filename: str) -> 'Analyzer':
        """
        Load the :py:class:`Analyzer` object from an HDF5 file.

        :param filename: Name for the file, any file extension will be replaced with .hdf.

        :return: An :py:class:`Analyzer` object from the file.
        """
        if filename[-4:] != '.hdf':
            raise ValueError("The expected file extension is .hdf")
        try:
            import h5py
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError("To save and load objects, the h5py module is required")  # pragma: no cover
        with h5py.File(filename, 'r') as h5file:
            my_dict = _group_to_dict(h5file, '/')
        return cls.from_dict(my_dict)

    def to_dict(self) -> dict:
        """
        :return: Dictionary description of Analyzer.
        """
        return {'delta_t': self._delta_t, 'disp_3d': self._disp_3d, 'volume': self._volume}

    @classmethod
    def from_dict(cls, my_dict: dict) -> 'Analyzer':
        """
        Generate an :py:class:`Analyzer` object from a dictionary.

        :param my_dict: The input dictionary.

        :return: New :py:class:`Analyzer` object.
        """
        return cls(**my_dict)

    @classmethod
    def _from_pymatgen(cls,
                       trajectory: List[Union['pymatgen.core.structure.Structure',
                                              List['pymatgen.core.structure.Structure']]],
                       parser_params: dict,
                       dtype: str = None,
                       **kwargs):
        """
        Create a :py:class:`Analyzer` object from a list or nested list of
        :py:class:`pymatgen.core.structure.Structure` objects.

        :param trajectory: The list or nested list of structures to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object.
            See the appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a list of :py:class:`pymatgen.core.structure.Structure`
            objects, this should be :py:attr:`None`. However, if a list of lists is passed, then it is necessary
            to identify if these constitute a series of :py:attr:`consecutive` trajectories or a series of
            :py:attr:`identical` starting points with different random seeds, in which case the `dtype` should
            be either :py:attr:`consecutive` or :py:attr:`identical`.

        :return: Relevant :py:class:`Analyzer` object.
        """
        if dtype is None:
            u = PymatgenParser(trajectory, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        elif dtype == 'identical':
            u = [PymatgenParser(f, **parser_params) for f in trajectory]
            return cls(u[0].delta_t, cls._stack_trajectories(u), u[0].volume, **kwargs)
        elif dtype == 'consecutive':
            structures = _flatten_list([x for x in trajectory])
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        else:
            raise ValueError('The dtype specified was not recognised, please consult the kinisi documentation.')

    @classmethod
    def _from_Xdatcar(cls,
                      trajectory: Union['pymatgen.io.vasp.outputs.Xdatcar', List['pymatgen.io.vasp.outputs.Xdatcar']],
                      parser_params: dict,
                      dtype: str = None,
                      **kwargs):
        """
        Create a :py:class:`Analyzer` object from a single or a list of
        :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects.

        :param trajectory: The Xdatcar or list of Xdatcar objects to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a :py:class:`pymatgen.io.vasp.outputs.Xdatcar` object, this
            should be :py:attr:`None`. However, if a list of :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects is
            passed, then it is necessary to identify if these constitute a series of :py:attr:`consecutive`
            trajectories or a series of :py:attr:`identical` starting points with different random seeds, in which
            case the `dtype` should be either :py:attr:`consecutive` or :py:attr:`identical`.

        :return: Relevant :py:class:`Analyzer` object.
        """
        if dtype is None:
            structures = trajectory.structures
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        elif dtype == 'identical':
            u = [PymatgenParser(f.structures, **parser_params) for f in trajectory]
            return cls(u[0].delta_t, cls._stack_trajectories(u), u[0].volume, **kwargs)
        elif dtype == 'consecutive':
            structures = _flatten_list([x.structures for x in trajectory])
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        else:
            raise ValueError('The dtype specified was not recognised, please consult the kinisi documentation.')

    @classmethod
    def _from_file(cls, trajectory: Union[str, List[str]], parser_params: dict, dtype: str = None, **kwargs):
        """
        Create a :py:class:`Analyzer` object from a single or a list of Xdatcar file(s).

        :param trajectory: The file or list of Xdatcar files to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify if these constitute a series of
            :py:attr:`consecutive` trajectories or a series of :py:attr:`identical` starting points with different
            random seeds, in which case the `dtype` should be either :py:attr:`consecutive` or :py:attr:`identical`.

        :return: Relevant :py:class:`Analyzer` object.
        """
        try:
            from pymatgen.io.vasp import Xdatcar
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError("To use the from_file method, pymatgen must be installed.")  # pragma: no cover
        if dtype is None:
            structures = Xdatcar(trajectory).structures
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        elif dtype == 'identical':
            u = [PymatgenParser(Xdatcar(f).structures, **parser_params) for f in trajectory]
            return cls(u[0].delta_t, cls._stack_trajectories(u), u[0].volume, **kwargs)
        elif dtype == 'consecutive':
            trajectory_list = (Xdatcar(f) for f in trajectory)
            structures = _flatten_list([x.structures for x in trajectory_list])
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        else:
            raise ValueError('The dtype specified was not recognised, please consult the kinisi documentation.')

    @classmethod
    def _from_universe(cls,
                       trajectory: 'MDAnalysis.core.universe.Universe',
                       parser_params: dict,
                       dtype: str = None,
                       **kwargs):
        """
        Create an :py:class:`Analyzer` object from an :py:class:`MDAnalysis.core.universe.Universe` object.

        :param trajectory: The universe to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.MDAnalysisParser` object.
            See the appropriate documention for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify that these constitute a series of
            :py:attr:`identical` starting points with different random seeds, in which case the `dtype` should
            be :py:attr:`identical`. For a series of consecutive trajectories, please construct the relevant
            object using :py:mod:`MDAnalysis`.

        :return: Relevant :py:class:`Analyzer` object.
        """
        try:
            import MDAnalysis as mda
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError(
                "To use the MDAnalysis from file parsing, MDAnalysis must be installed.")  # pragma: no cover
        if dtype is None:
            u = MDAnalysisParser(trajectory, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        elif dtype == 'identical':
            u = [MDAnalysisParser(t, **parser_params) for t in trajectory]
            return cls(u[0].delta_t, cls._stack_trajectories(u), u[0].volume, **kwargs)
        else:
            raise ValueError('The dtype specified was not recognised, please consult the kinisi documentation.')

    @staticmethod
    def _stack_trajectories(u: Union[MDAnalysisParser, PymatgenParser]) -> List[np.ndarray]:
        """
        If more than one trajectory is given, then they are stacked to give the appearence that there are
        additional atoms in the trajectory.

        :param u: Results from the parsing of each trajectory.

        :return: The stacked displacement list.
        """
        joint_disp_3d = []
        for i in range(len(u[0].disp_3d)):
            disp = np.zeros((u[0].disp_3d[i].shape[0] * len(u), u[0].disp_3d[i].shape[1], u[0].disp_3d[i].shape[2]))
            disp[:u[0].disp_3d[i].shape[0]] = u[0].disp_3d[i]
            for j in range(1, len(u)):
                disp[u[0].disp_3d[i].shape[0] * j:u[0].disp_3d[i].shape[0] * (j + 1)] = u[j].disp_3d[i]
            joint_disp_3d.append(disp)
        return joint_disp_3d
    
    @property
    def distribution(self) -> np.ndarray:
        """
        :return: A distribution of samples for the linear relationship that can be used for easy
        plotting of credible intervals.
        """
        return self._diff.gradient.samples * self.dt[:, np.newaxis] + self._diff.intercept.samples

    @property
    def dt(self) -> np.ndarray:
        """
        :return: Timestep values that have been sampled.
        """
        return self._diff.dt

    @property
    def dr(self) -> List['uravu.distribution.Distribution']:
        """
        :return: A list of :py:class:`uravu.distribution.Distribution` objects that describe the euclidian
            displacement at each :py:attr:`dt`.
        """
        return self._diff.euclidian_displacements

    @property
    def ngp_max(self) -> float:
        """
        :return: Position in dt where the non-Gaussian parameter is maximised.
        """
        return self.dt[self._diff.ngp.argmax()]

    @property
    def intercept(self) -> 'uravu.distribution.Distribution':
        """
        :return: The distribution describing the intercept.
        """
        return self._diff.intercept

    @property
    def volume(self) -> float:
        """
        :return: Volume of system, in cubic angstrom.
        """
        return self._volume


def _flatten_list(this_list: list) -> list:
    """
    Flatten nested lists.

    :param this_list: List to be flattened

    :return: Flattened list
    """
    return [item for sublist in this_list for item in sublist]


def _dict_to_group(h5file: 'h5py._hl.files.File', path: str, my_dict: dict):
    """
    A recursive function to help with saving to hdf5 file formats.

    :param h5file: Open hdf5 file.
    :param path: Path in the hdf5 file.
    :param my_dict: Dict to make datasets from.
    """
    for key, value in my_dict.items():
        if isinstance(value, (np.ndarray, int, float, str, bytes)):
            h5file[path + key] = value
        elif isinstance(value, list):
            for i, d in enumerate(value):
                if isinstance(d, (np.ndarray, int, float, str, bytes)):
                    h5file[path + key + f'/list{i}'] = d
                elif isinstance(d, dict):
                    _dict_to_group(h5file, path + key + f'/list{i}' + '/', d)
        elif isinstance(value, dict):
            _dict_to_group(h5file, path + key + '/', value)
        elif value is None:
            h5file[path + key] = 'NULL'
        else:
            raise ValueError(f'Cannot save {type(value)} type')


def _group_to_dict(h5file: 'h5py._hl.files.File', path: str) -> dict:
    """
    A recursive function to load data from hdf5 files.

    :param h5file: Open hdf5 file.
    :param path: Path in the hdf5 file.

    :return: A dictionary of the hdf5 groups and datasets.
    """
    import h5py
    my_dict = {}
    for key, value in h5file[path].items():
        if isinstance(value, h5py._hl.dataset.Dataset):
            if isinstance(value[()], (str, bytes)):
                if value[()] == b'NULL':
                    my_dict[key] = None
                else:
                    my_dict[key] = value[()]
            else:
                my_dict[key] = value[()]
        elif isinstance(value, h5py._hl.group.Group):
            key_list = list(h5file[path + key + '/'].keys())
            if key_list[0][:4] == 'list':
                my_dict[key] = []
                for i in sorted([int(i[4:]) for i in key_list]):
                    value = h5file[path + key + f'/list{i}']
                    if isinstance(value, h5py._hl.dataset.Dataset):
                        if isinstance(value[()], (str, bytes)):
                            if value[()] == b'NULL':
                                my_dict[key].append(None)
                            else:
                                my_dict[key].append(value[()])
                        else:
                            my_dict[key].append(value[()])
                    elif isinstance(h5file[path + key + f'/list{i}'], h5py._hl.group.Group):
                        my_dict[key].append(_group_to_dict(h5file, path + key + f'/list{i}'))
            else:
                my_dict[key] = _group_to_dict(h5file, path + key + '/')
        else:
            raise ValueError(f'Cannot save {type(value)} type')
    return my_dict



"""
The modules is focused on tools for the evaluation of the mean squared displacement and resulting
diffusion coefficient from a material.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (arm61)

import warnings
from typing import List, Tuple, Union
import numpy as np
from scipy.stats import multivariate_normal, normaltest, linregress
from scipy.optimize import minimize, curve_fit
import scipy.constants as const
import tqdm
from uravu.distribution import Distribution
from sklearn.utils import resample
from emcee import EnsembleSampler
from kinisi.matrix import find_nearest_positive_definite

DIMENSIONALITY = {
    'x': np.s_[0],
    'y': np.s_[1],
    'z': np.s_[2],
    'xy': np.s_[:2],
    'xz': np.s_[::2],
    'yz': np.s_[1:],
    'xyz': np.s_[:]
}


class Bootstrap:
    """
    The top-level class for bootstrapping.

    :param delta_t: An array of the timestep values.
    :param disp_3d: A list of arrays, where each array has the axes [atom, displacement observation, dimension].
        There is one array in the list for each delta_t value. Note: it is necessary to use a list of arrays as
        the number of observations is not necessary the same at each data point.
    :param sub_sample_dt: The frequency in observations to be sampled. Optional, default is :py:attr:`1` (every
        observation).
    :param progress: Show tqdm progress for sampling. Optional, default is :py:attr:`True`.
    """

    def __init__(self, delta_t: np.ndarray, disp_3d: List[np.ndarray], sub_sample_dt: int = 1, progress: bool = True):
        self._displacements = disp_3d[::sub_sample_dt]
        self._delta_t = np.array(delta_t[::sub_sample_dt])
        self._max_obs = self._displacements[0].shape[1]
        self._distributions = []
        self._dt = np.array([])
        self._iterator = self.iterator(progress, range(len(self._displacements))[1:])
        self._n = np.array([])
        self._s = np.array([])
        self._v = np.array([])
        self._n_i = np.array([], dtype=int)
        self._ngp = np.array([])
        self._euclidian_displacements = []
        self._diffusion_coefficient = None
        self._jump_diffusion_coefficient = None
        self._sigma = None
        self._intercept = None
        self.gradient = None
        self.flatchain = None
        self._covariance_matrix = None
        self.dims = 3

    def to_dict(self) -> dict:
        """
        :return: Dictionary description of the :py:class:`Bootstrap`.
        """
        my_dict = {
            'displacements': self._displacements,
            'delta_t': self._delta_t,
            'max_obs': self._max_obs,
            'dt': self._dt,
            'n': self._n,
            's': self._s,
            'v': self._v,
            'n_i': self._n_i,
            'ngp': self._ngp,
            'covariance_matrix': self._covariance_matrix,
            'diffusion_coefficient': None,
            'jump_diffusion_coefficient': None,
            'sigma': None,
            'intercept': None,
            'gradient': None
        }
        my_dict['distributions'] = [d.to_dict() for d in self._distributions]
        my_dict['euclidian_displacements'] = [d.to_dict() for d in self._euclidian_displacements]
        if self._diffusion_coefficient is not None:
            my_dict['diffusion_coefficient'] = self._diffusion_coefficient.to_dict()
        if self._jump_diffusion_coefficient is not None:
            my_dict['jump_diffusion_coefficient'] = self._jump_diffusion_coefficient.to_dict()
        if self._sigma is not None:
            my_dict['sigma'] = self._sigma.to_dict()
        if self._intercept is not None:
            my_dict['intercept'] = self._intercept.to_dict()
        if self.gradient is not None:
            my_dict['gradient'] = self.gradient.to_dict()
        my_dict['flatchain'] = self.flatchain
        return my_dict

    @classmethod
    def from_dict(cls, my_dict: dict) -> 'Bootstrap':
        """
        Generate a :py:class:`Bootstrap` object from a dictionary.

        :param my_dict: The input dictionary.

        :return: New :py:class`Bootstrap` object.
        """
        boot = cls(my_dict['delta_t'], my_dict['displacements'], sub_sample_dt=1, progress=False)
        boot._max_obs = my_dict['max_obs']
        boot._distributions = [Distribution.from_dict(d) for d in my_dict['distributions']]
        boot._euclidian_displacements = [Distribution.from_dict(d) for d in my_dict['euclidian_displacements']]
        boot._dt = my_dict['dt']
        boot._n = my_dict['n']
        boot._s = my_dict['s']
        boot._v = my_dict['v']
        boot._n_i = my_dict['n_i']
        boot._ngp = my_dict['ngp']
        if my_dict['diffusion_coefficient'] is not None:
            boot._diffusion_coefficient = Distribution.from_dict(my_dict['diffusion_coefficient'])
        if my_dict['jump_diffusion_coefficient'] is not None:
            boot._jump_diffusion_coefficient = Distribution.from_dict(my_dict['jump_diffusion_coefficient'])
        if my_dict['sigma'] is not None:
            boot._sigma = Distribution.from_dict(my_dict['sigma'])
        if my_dict['intercept'] is not None:
            boot._intercept = Distribution.from_dict(my_dict['intercept'])
        if my_dict['gradient'] is not None:
            boot.gradient = Distribution.from_dict(my_dict['gradient'])
        boot.flatchain = my_dict['flatchain']
        boot._covariance_matrix = my_dict['covariance_matrix']
        return boot

    @property
    def dt(self) -> np.ndarray:
        """
        :return: Timestep values that were resampled.
        """
        return self._dt

    @property
    def n(self) -> np.ndarray:
        """
        :return: The mean MSD/TMSD/MSCD, as determined from the bootstrap resampling process, in units Å:sup:`2`.
        """
        return self._n

    @property
    def s(self) -> np.ndarray:
        """
        :return: The MSD/TMSD/MSCD standard deviation, as determined from the bootstrap resampling process, in
            units Å:sup:`2`.
        """
        return self._s

    @property
    def v(self) -> np.ndarray:
        """
        :return: The MSD/TMSD/MSCD variance as determined from the bootstrap resampling process, in units Å:sup:`4`.
        """
        return self._v

    @property
    def euclidian_displacements(self) -> List[Distribution]:
        """
        :return: Displacements between particles at each dt.
        """
        return self._euclidian_displacements

    @property
    def ngp(self) -> np.ndarray:
        """
        :return: Non-Gaussian parameter as a function of :py:attr:`dt`.
        """
        return self._ngp

    @property
    def n_i(self) -> np.ndarray:
        """
        :return: The number of independent trajectories as a function of :py:attr:`dt`.
        """
        return self._n_i

    @property
    def intercept(self) -> Union[Distribution, None]:
        """
        :return: The estimated intercept. Note that if :py:attr:`fit_intercept` is :py:attr:`False` is the
            relavent method call, then this is :py:attr:`None`
        """
        return self._intercept

    @property
    def covariance_matrix(self) -> np.ndarray:
        """
        :return: The covariance matrix for the trajectories.
        """
        return self._covariance_matrix

    @staticmethod
    def iterator(progress: bool, loop: Union[list, range]) -> Union[tqdm.std.tqdm, range]:
        """
        Get the iteration object, using :py:mod:`tqdm` as appropriate.

        :param progress: Should :py:mod:`tqdm` be used to give a progress bar.
        :param loop: The object that should be looped over.

        :return: Iterator object.
        """
        if progress:
            return tqdm.tqdm(loop, desc='Bootstrapping Displacements')
        return loop

    @staticmethod
    def sample_until_normal(array: np.ndarray,
                            n_samples: int,
                            n_resamples: int,
                            max_resamples: int,
                            alpha: float = 1e-3,
                            random_state: np.random.mtrand.RandomState = None) -> Distribution:
        """
        Resample from the distribution until a normal distribution is obtained or a maximum is reached.

        Args:
        :param array: The array to sample from.
        :param n_samples: Number of samples.
        :param r_resamples: Number of resamples to perform initially.
        :param max_resamples: The maximum number of resamples to perform.
        :param alpha: Level that p-value should be below in :py:func:`scipy.stats.normaltest` for the distribution
            to be normal. Optional, default is :py:attr:`1e-3`.
        :param random_state: A :py:attr:`RandomState` object to be used to ensure reproducibility. Optional,
            default is :py:attr:`None`.

        :return: The resampled distribution.
        """
        values = _bootstrap(array.flatten(), n_samples, n_resamples, random_state)
        p_value = normaltest(values)[1]
        while p_value < alpha and len(values) < max_resamples:
            values += _bootstrap(array.flatten(), n_samples, 100, random_state)
            p_value = normaltest(values)[1]
        if len(values) >= max_resamples:
            warnings.warn("The maximum number of resamples has been reached, and the distribution is not yet normal.")
        return Distribution(values)

    @staticmethod
    def n_samples(disp_shape: Tuple[float, float], max_obs: int) -> int:
        """
        Calculate the maximum number of independent observations.

        :param disp_shape:: The shape of the displacements array.
        :param max_obs: The maximum number of observations for the trajectory.

        :return: Maximum number of independent observations.
        """
        n_obs = disp_shape[1]
        n_atoms = disp_shape[0]
        dt_int = max_obs - n_obs + 1
        return int(max_obs / dt_int * n_atoms)

    @staticmethod
    def ngp_calculation(d_squared: np.ndarray) -> float:
        """
        Determine the non-Gaussian parameter, from S. Song et al, "Transport dynamics of complex fluids"
        (2019): 10.1073/pnas.1900239116

        :param d_squared: Squared displacement values.

        :return: Value of non-Gaussian parameter.
        """
        top = np.mean(d_squared * d_squared) * 3
        bottom = np.square(np.mean(d_squared)) * 5
        return top / bottom - 1

    def bootstrap_GLS(self,
                      use_ngp: bool = False,
                      dt_skip: float = 0,
                      fit_intercept: bool = True,
                      n_samples: int = 1000,
                      n_walkers: int = 32,
                      n_burn: int = 500,
                      progress: bool = True,
                      random_state: np.random.mtrand.RandomState = None):
        """
        Use the covariance matrix estimated from the resampled values to estimate the gradient and intercept
        using a generalised least squares approach.

        :param use_ngp: Should the ngp max be used as the starting point for the diffusion fitting. Optional,
            default is :py:attr:`False`.
        :param dt_skip: Values of :py:attr:`dt` that should be skipped, i.e. where the atoms are not diffusing.
            Note that if :py:attr:`use_ngp` is :py:attr:`True` this will be ignored. Optional, defaults
            to :py:attr:`0`.
        :param fit_intercept: Should the intercept of the diffusion relationship be fit. Optional, default
            is :py:attr:`True`.
        :param n_samples: Number of samples of the Gaussian process to perform. Optional, default is :py:attr:`1000`.
        :param n_walkers: Number of MCMC walkers to use. Optional, default is :py:attr:`32`.
        :param n_burn: Number of burn in samples (these allow the sampling to settle). Optional, default
            is :py:attr:`500`.
        :param rtol: The relative threshold term for the covariance matrix inversion. If you obtain a very unusual
            value for the diffusion coefficient, it is recommended to increase this value (ideally iteratively).
            Optional, default is :code:`N * eps`, where :code:`eps` is the machine precision value of the covariance
            matrix content.
        :param progress: Show tqdm progress for sampling. Optional, default is :py:attr:`True`.
        :param random_state: A :py:attr:`RandomState` object to be used to ensure reproducibility. Optional,
            default is :py:attr:`None`.
        """
        max_ngp = np.argwhere(self._dt > dt_skip)[0][0]
        if use_ngp:
            max_ngp = np.argmax(self._ngp)

        def model_variance(dt: np.ndarray, a: float) -> np.ndarray:
            """
            Determine the model variance, based on a quadratic relationship with the number of
            independent samples as a divisor.

            :param dt: Timestep value
            :param a: Quadratic coefficient
            :return: Model variances
            """
            return a / self._n_i[max_ngp:] * dt**2

        popt, _ = curve_fit(model_variance, self.dt[max_ngp:], self._v[max_ngp:])
        model_v = model_variance(self.dt[max_ngp:], *popt)

        self._covariance_matrix = self.populate_covariance_matrix(model_v, self._n_i)
        self._covariance_matrix = find_nearest_positive_definite(self._covariance_matrix)

        mv = multivariate_normal(self._n[max_ngp:], self._covariance_matrix, allow_singular=True)

        def log_likelihood(theta: np.ndarray) -> float:
            """
            Get the log likelihood for multivariate normal distribution.
            :param theta: Value of the gradient and intercept of the straight line.
            :return: Log-likelihood value.
            """
            if theta[0] < 0:
                return -np.inf
            model = _straight_line(self._dt[max_ngp:], *theta)
            return mv.logpdf(model)

        ols = linregress(self._dt[max_ngp:], self._n[max_ngp:])
        slope = ols.slope
        intercept = 1e-20
        if slope < 0:
            slope = 1e-20

        def nll(*args) -> float:
            """
            General purpose negative log-likelihood.
            :return: Negative log-likelihood
            """
            return -log_likelihood(*args)

        if fit_intercept:
            max_likelihood = minimize(nll, np.array([slope, intercept])).x
        else:
            max_likelihood = minimize(nll, np.array([slope])).x
        pos = max_likelihood + max_likelihood * 1e-3 * np.random.randn(n_walkers, max_likelihood.size)
        sampler = EnsembleSampler(*pos.shape, log_likelihood)
        # Waiting on https://github.com/dfm/emcee/pull/376
        # if random_state is not None:
        #     pos = max_likelihood + max_likelihood * 1e-3 * random_state.randn(n_walkers, max_likelihood.size)
        #     sampler._random = random_state
        sampler.run_mcmc(pos, n_samples + n_burn, progress=progress, progress_kwargs={'desc': "Likelihood Sampling"})
        self.flatchain = sampler.get_chain(flat=True, discard=n_burn)

        self.gradient = Distribution(self.flatchain[:, 0])
        self._intercept = None
        if fit_intercept:
            self._intercept = Distribution(self.flatchain[:, 1])

    @staticmethod
    def populate_covariance_matrix(variances: np.ndarray, n_samples: np.ndarray) -> np.ndarray:
        """
        Populate the covariance matrix for the generalised least squares methodology.

        :param variances: The variances for each timestep
        :param n_samples: Number of independent trajectories for each timestep

        :return: An estimated covariance matrix for the system
        """
        covariance_matrix = np.zeros((variances.size, variances.size))
        for i in range(0, variances.size):
            for j in range(i, variances.size):
                ratio = n_samples[i] / n_samples[j]
                value = ratio * variances[i]
                covariance_matrix[i, j] = value
                covariance_matrix[j, i] = np.copy(covariance_matrix[i, j])
        return covariance_matrix

    def diffusion(self, **kwargs):
        """
        Use the bootstrap-GLS method to determine the diffusivity for the system. Keyword arguments will be
        passed of the :py:func:`bootstrap_GLS` method.
        """
        self.bootstrap_GLS(**kwargs)
        self._diffusion_coefficient = Distribution(self.gradient.samples / (2e4 * self.dims))

    @property
    def D(self) -> Union[Distribution, None]:
        """
        An alias for the diffusion coefficient Distribution.

        :return: Diffusion coefficient, with units of cm:sup:`2`s:sup:`-1`.
        """
        return self._diffusion_coefficient

    def jump_diffusion(self, **kwargs):
        """
        Use the bootstrap-GLS method to determine the jump diffusivity for the system. Keyword arguments
        will be passed of the :py:func:`bootstrap_GLS` method.
        """
        self.bootstrap_GLS(**kwargs)
        self._jump_diffusion_coefficient = Distribution(
            self.gradient.samples / (2e4 * self.dims * self._displacements[0].shape[0]))

    @property
    def D_J(self) -> Union[Distribution, None]:
        """
        Alias for the jump diffusion coefficient Distribution.

        :return: Jump diffusion coefficient, with units of cm:sup:`2`s:sup:`-1`.
        """
        return self._jump_diffusion_coefficient

    def conductivity(self, temperature: float, volume: float, **kwargs):
        """
        Use the bootstrap-GLS method to determine the ionic conductivity for the system, in units of mScm:sup:`-1`.
        Keyword arguments will be passed of the :py:func:`bootstrap_GLS` method.

        :param temperature: System temperature, in Kelvin.
        :param volume: System volume, in Å^{3}.
        """
        self.bootstrap_GLS(**kwargs)
        volume = volume * 1e-24  # cm^3
        D = self.gradient.samples / (2e4 * self.dims)  # cm^2s^-1
        conversion = 1000 / (volume * const.N_A) * (const.N_A * const.e)**2 / (const.R * temperature)
        self._sigma = Distribution(D * conversion)

    @property
    def sigma(self) -> Union[Distribution, None]:
        """
        :return: The estimated conductivity, based on the generalised least squares approach, with
            units mScm:sup:`-1`.
        """
        return self._sigma


class MSDBootstrap(Bootstrap):
    """
    Perform a bootstrap resampling to obtain accurate estimates for the mean and uncertainty for the mean
    squared displacements.

    :param delta_t: An array of the timestep values, units of ps
    :param disp_3d: A list of arrays, where each array has the axes
        :code:`[atom, displacement observation, dimension]`. There is one array in the list for each
        delta_t value. Note: it is necessary to use a list of arrays as the number of observations is
        not necessary the same at each data point.
    :param sub_sample_dt: The frequency in observations to be sampled. Default is :py:attr:`1` (every observation)
    :param n_resamples: The initial number of resamples to be performed. Default is :py:attr:`1000`
    :param max_resamples: The max number of resamples to be performed by the distribution is assumed to be normal.
        This is present to allow user control over the time taken for the resampling to occur. Default
        is :py:attr:`100000`
    :param dimension: Dimension/s to find the displacement along, this should be some subset of `'xyz'` indicating
        the axes of interest. Optional, defaults to `'xyz'`.
    :param alpha: Value that p-value for the normal test must be greater than to accept. Default is :py:attr:`1e-3`
    :param random_state : A :py:attr:`RandomState` object to be used to ensure reproducibility. Default
        is :py:attr:`None`
    :param progress: Show tqdm progress for sampling. Default is :py:attr:`True`
    """

    def __init__(self,
                 delta_t: np.ndarray,
                 disp_3d: List[np.ndarray],
                 sub_sample_dt: int = 1,
                 n_resamples: int = 1000,
                 max_resamples: int = 10000,
                 dimension: str = 'xyz',
                 alpha: float = 1e-3,
                 random_state: np.random.mtrand.RandomState = None,
                 progress: bool = True):
        super().__init__(delta_t, disp_3d, sub_sample_dt, progress)
        slice = DIMENSIONALITY[dimension.lower()]
        self.dims = len(dimension.lower())
        for i in self._iterator:
            disp_slice = self._displacements[i][:, :, slice].reshape(self._displacements[i].shape[0],
                                                                     self._displacements[i].shape[1], self.dims)
            d_squared = np.sum(disp_slice**2, axis=2)
            n_samples_current = self.n_samples(self._displacements[i].shape, self._max_obs)
            if n_samples_current <= 1:
                continue
            self._n_i = np.append(self._n_i, n_samples_current)
            self._euclidian_displacements.append(Distribution(np.sqrt(d_squared.flatten())))
            distro = self.sample_until_normal(d_squared, n_samples_current, n_resamples, max_resamples, alpha,
                                              random_state)
            self._distributions.append(distro)
            self._n = np.append(self._n, d_squared.mean())
            self._s = np.append(self._s, np.std(distro.samples, ddof=1))
            self._v = np.append(self._v, np.var(distro.samples, ddof=1))
            self._ngp = np.append(self._ngp, self.ngp_calculation(d_squared))
            self._dt = np.append(self._dt, self._delta_t[i])


class TMSDBootstrap(Bootstrap):
    """
    Perform a bootstrap resampling to obtain accurate estimates for the mean and uncertainty for the total
    mean squared displacements.

    :param delta_t: An array of the timestep values.
    :param disp_3d: A list of arrays, where each array has the axes
        :code:`[atom, displacement observation, dimension]`. There is one array in the list for each
        delta_t value. Note: it is necessary to use a list of arrays as the number of observations is
        not necessary the same at each data point.
    :param sub_sample_dt: The frequency in observations to be sampled. Optional, default
        is :py:attr:`1` (every observation)
    :param n_resamples: The initial number of resamples to be performed. Optional, default
        is :py:attr:`1000`
    :param max_resamples: The max number of resamples to be performed by the distribution is assumed to be
        normal. This is present to allow user control over the time taken for the resampling to occur.
        Optional, default is :py:attr:`100000`
    :param dimension: Dimension/s to find the displacement along, this should be some subset of `'xyz'` indicating
        the axes of interest. Optional, defaults to `'xyz'`.
    :param alpha: Value that p-value for the normal test must be greater than to accept. Optional, default
        is :py:attr:`1e-3`
    :param random_state : A :py:attr:`RandomState` object to be used to ensure reproducibility. Optional,
        default is :py:attr:`None`
    :param progress: Show tqdm progress for sampling. Optional, default is :py:attr:`True`
    """

    def __init__(self,
                 delta_t: np.ndarray,
                 disp_3d: List[np.ndarray],
                 sub_sample_dt: int = 1,
                 n_resamples: int = 1000,
                 max_resamples: int = 10000,
                 dimension: str = 'xyz',
                 alpha: float = 1e-3,
                 random_state: np.random.mtrand.RandomState = None,
                 progress: bool = True):
        super().__init__(delta_t, disp_3d, sub_sample_dt, progress)
        slice = DIMENSIONALITY[dimension.lower()]
        self.dims = len(dimension.lower())
        for i in self._iterator:
            disp_slice = self._displacements[i][:, :, slice].reshape(self._displacements[i].shape[0],
                                                                     self._displacements[i].shape[1], self.dims)
            d_squared = np.sum(disp_slice**2, axis=2)
            coll_motion = np.sum(np.sum(disp_slice, axis=0)**2, axis=-1)
            n_samples_current = self.n_samples((1, self._displacements[i].shape[1]), self._max_obs)
            if n_samples_current <= 1:
                continue
            self._n_i = np.append(self._n_i, n_samples_current)
            self._euclidian_displacements.append(Distribution(np.sqrt(d_squared.flatten())))
            distro = self.sample_until_normal(coll_motion, n_samples_current, n_resamples, max_resamples, alpha,
                                              random_state)
            self._distributions.append(distro)
            self._n = np.append(self._n, distro.n)
            self._s = np.append(self._s, np.std(distro.samples, ddof=1))
            self._v = np.append(self._v, np.var(distro.samples, ddof=1))
            self._ngp = np.append(self._ngp, self.ngp_calculation(d_squared.flatten()))
            self._dt = np.append(self._dt, self._delta_t[i])


class MSCDBootstrap(Bootstrap):
    """
    Perform a bootstrap resampling to obtain accurate estimates for the mean and uncertainty for the mean
    squared charge displacements.

    :param delta_t: An array of the timestep values.
    :param disp_3d: A list of arrays, where each array has the axes
        :code:`[atom, displacement observation, dimension]`. There is one array in the list for each
        delta_t value. Note: it is necessary to use a list of arrays as the number of observations is
        not necessary the same at each data point.
    :param ionic_charge: The charge on the mobile ions, either an array with a value for each ion or a scalar
        if all values are the same.
    :param sub_sample_dt: The frequency in observations to be sampled. Optional, default is :py:attr:`1`
        (every observation).
    :param n_resamples: The initial number of resamples to be performed. Optional, default is :py:attr:`1000`.
    :param max_resamples: The max number of resamples to be performed by the distribution is assumed to be normal.
        This is present to allow user control over the time taken for the resampling to occur. Optional, default
        is :py:attr:`100000`.
    :param dimension: Dimension/s to find the displacement along, this should be some subset of `'xyz'` indicating
        the axes of interest. Optional, defaults to `'xyz'`.
    :param alpha: Value that p-value for the normal test must be greater than to accept. Optional, default
        is :py:attr:`1e-3`.
    :param random_state: A :py:attr:`RandomState` object to be used to ensure reproducibility. Optional,
        default is :py:attr:`None`.
    :param progress: Show tqdm progress for sampling. Optional, default is :py:attr:`True`.
    """

    def __init__(self,
                 delta_t: np.ndarray,
                 disp_3d: List[np.ndarray],
                 ionic_charge: Union[np.ndarray, int],
                 sub_sample_dt: int = 1,
                 n_resamples: int = 1000,
                 max_resamples: int = 10000,
                 dimension: str = 'xyz',
                 alpha: float = 1e-3,
                 random_state: np.random.mtrand.RandomState = None,
                 progress: bool = True):
        super().__init__(delta_t, disp_3d, sub_sample_dt, progress)
        try:
            _ = len(ionic_charge)
        except TypeError:
            ionic_charge = np.ones(self._displacements[0].shape[0]) * ionic_charge
        slice = DIMENSIONALITY[dimension.lower()]
        self.dims = len(dimension.lower())
        for i in self._iterator:
            disp_slice = self._displacements[i][:, :, slice].reshape(self._displacements[i].shape[0],
                                                                     self._displacements[i].shape[1], self.dims)
            d_squared = np.sum(disp_slice**2, axis=2)
            sq_chg_motion = np.sum(np.sum((ionic_charge * self._displacements[i].T).T, axis=0)**2, axis=-1)
            n_samples_current = self.n_samples((1, self._displacements[i].shape[1]), self._max_obs)
            if n_samples_current <= 1:
                continue
            self._n_i = np.append(self._n_i, n_samples_current)
            self._euclidian_displacements.append(Distribution(np.sqrt(d_squared.flatten())))
            distro = self.sample_until_normal(sq_chg_motion, n_samples_current, n_resamples, max_resamples, alpha,
                                              random_state)
            self._distributions.append(distro)
            self._n = np.append(self._n, distro.n)
            self._s = np.append(self._s, np.std(distro.samples, ddof=1))
            self._v = np.append(self._v, np.var(distro.samples, ddof=1))
            self._ngp = np.append(self._ngp, self.ngp_calculation(d_squared.flatten()))
            self._dt = np.append(self._dt, self._delta_t[i])


def _bootstrap(array: np.ndarray, n_samples: int, n_resamples: int, random_state: np.random.mtrand.RandomState = None):
    """
    Perform a set of resamples.

    :param array: The array to sample from.
    :param n_samples: Number of samples.
    :param n_resamples: Number of resamples to perform initially.
    :param random_state: A :py:attr:`RandomState` object to be used to ensure reproducibility. Optional,
        default is :py:attr:`None`

    :return: Resampled values from the array
    """
    return [
        np.mean(resample(array.flatten(), n_samples=n_samples, random_state=random_state)) for j in range(n_resamples)
    ]


def _straight_line(abscissa: np.ndarray, gradient: float, intercept: float = 0.0) -> np.ndarray:
    """
    A one dimensional straight line function.

    :param abscissa: The abscissa data.
    :param gradient: The slope of the line.
    :param intercept: The y-intercept of the line. Optional, default is :py:attr:`0.0`.

    :return: The resulting ordinate.
    """
    return gradient * abscissa + intercept
